package com.tzld.piaoquan.abtest.service;

import com.ctrip.framework.apollo.spring.annotation.ApolloJsonValue;
import com.tzld.piaoquan.abtest.client.model.GetGroupRequest;
import com.tzld.piaoquan.abtest.client.model.GetGroupResponse;
import com.tzld.piaoquan.abtest.client.model.GroupInfoProto;
import com.tzld.piaoquan.abtest.client.model.Result;
import com.tzld.piaoquan.abtest.model.ExpConfig;
import com.tzld.piaoquan.abtest.util.CommonCollectionUtils;
import lombok.extern.slf4j.Slf4j;
import org.apache.commons.collections4.CollectionUtils;
import org.apache.commons.collections4.MapUtils;
import org.apache.commons.lang.RandomStringUtils;
import org.apache.commons.lang3.RandomUtils;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.beans.factory.annotation.Qualifier;
import org.springframework.data.redis.core.RedisTemplate;
import org.springframework.stereotype.Service;

import java.net.InetAddress;
import java.util.*;

/**
 * @author dyp
 */
@Service
@Slf4j
public class GroupService {

    @Autowired
    @Qualifier("redisTemplate")
    private RedisTemplate<String, String> redisTemplate;

    @ApolloJsonValue("${new.exp,config.v2:[]}")
    private List<ExpConfig> expConfigs;


    private final String scene_open = "open";
    private final String scene_daily_share_return = "daily_share_return";
    private final String scene_not_daily_share_return = "not_daily_share_return";


    public GetGroupResponse getGroup(GetGroupRequest request) {
        try {


            Map<String, List<ExpConfig.Layer>> layerMap = CommonCollectionUtils.toMap(expConfigs, c -> c.getAppType(),
                    c -> c.getLayers());
            List<ExpConfig.Layer> layers = layerMap.get(String.valueOf(request.getAppType()));

            if (CollectionUtils.isEmpty(layers)) {
                GetGroupResponse.Builder builder = GetGroupResponse.newBuilder();
                builder.setResult(Result.newBuilder().setCode(1));
                if (request.hasGroupInfo()) {
                    builder.setGroupInfo(request.getGroupInfo());
                }
                return builder.build();
            }

            //TODO 1. 确定场景 主动打开、当日分享、历史分享
            String scene = "";
            long time = request.hasGroupInfo()
                    ? request.getGroupInfo().getDate()
                    : 0;
            int isFromCard = request.getIsFromCard();

            if (isFromCard == 1) {
                if (isToday(time)) {
                    // 当日分享
                    scene = scene_daily_share_return;
                } else {
                    // 历史分享
                    scene = scene_not_daily_share_return;
                }
            } else {
                // 主动打开
                scene = scene_open;
            }

            log.info("ExperimentInterceptor scene {}", scene);

            //TODO 2. 根据场景 确定策略并分组
            Map<String, Integer> oldGroupMap = request.hasGroupInfo()
                    ? request.getGroupInfo().getGroupMap()
                    : Collections.emptyMap();
            Map<String, Integer> groupMap = new HashMap<>();
            for (ExpConfig.Layer layer : layers) {
                int bucketNum = layer.getBucketNum() <= 0 ? 100 : layer.getBucketNum();
                Map<String, String> groupRule = layer.getGroupRule();
                if (MapUtils.isEmpty(groupRule)) {
                    // random
                    groupMap.put(layer.getLayerId(), RandomUtils.nextInt(0, bucketNum));
                } else {
                    String policy = groupRule.get(scene);
                    if (org.apache.commons.lang.StringUtils.isBlank(policy)) {
                        // random
                        groupMap.put(layer.getLayerId(), RandomUtils.nextInt(0, bucketNum));
                    }
                    switch (policy) {

                        case "inherit":
                            groupMap.put(layer.getLayerId(), oldGroupMap.getOrDefault(layer.getLayerId(),
                                    RandomUtils.nextInt(0, bucketNum)));
                            break;
                        case "random_once":
                            if (!isToday(time)) {
                                groupMap.put(layer.getLayerId(), RandomUtils.nextInt(0, bucketNum));
                            } else {
                                groupMap.put(layer.getLayerId(), oldGroupMap.getOrDefault(layer.getLayerId(),
                                        RandomUtils.nextInt(0, bucketNum)));
                            }
                            break;
                        case "random_every":
                        default:
                            groupMap.put(layer.getLayerId(), RandomUtils.nextInt(0, bucketNum));
                            break;
                    }
                }
            }

            return GetGroupResponse.newBuilder()
                    .setResult(Result.newBuilder().setCode(1))
                    .setGroupInfo(GroupInfoProto.newBuilder()
                            .setDate(System.currentTimeMillis())
                            .setChain("")
                            .putAllGroup(groupMap))
                    .build();

        } catch (
                Exception e) {
            log.error("get group error", e);
            return GetGroupResponse.newBuilder()
                    .setResult(Result.newBuilder().setCode(0).setMessage("get group error"))
                    .build();
        }

    }


    public static String genChainId() {
        StringBuilder traceIdBuilder = new StringBuilder();
        InetAddress address = InetAddress.getLoopbackAddress();
        if (address == null) {
            return "default";
        }
        String addressStr = address.getHostAddress();
        if (org.apache.commons.lang3.StringUtils.isBlank(addressStr)) {
            return "default";
        }
        String[] addressArr = addressStr.split("\\.");
        if (addressArr.length != 4) {
            return "default";
        }

        traceIdBuilder.append(org.apache.commons.lang3.StringUtils.leftPad(Integer.toHexString(Integer.valueOf(addressArr[0])), 2, '0'))
                .append(org.apache.commons.lang3.StringUtils.leftPad(Integer.toHexString(Integer.valueOf(addressArr[1])), 2, '0'))
                .append(org.apache.commons.lang3.StringUtils.leftPad(Integer.toHexString(Integer.valueOf(addressArr[2])), 2, '0'))
                .append(org.apache.commons.lang3.StringUtils.leftPad(Integer.toHexString(Integer.valueOf(addressArr[3])), 2, '0'))
                .append(System.currentTimeMillis())
                .append(RandomStringUtils.randomAlphabetic(4))
                .append(Thread.currentThread().getId());
        return traceIdBuilder.toString();
    }

    private boolean isToday(long time) {
        Calendar calendar = Calendar.getInstance();
        calendar.setTimeInMillis(time);
        Calendar current = Calendar.getInstance();
        current.setTime(new Date());
        return calendar.get(Calendar.YEAR) == current.get(Calendar.YEAR)
                && calendar.get(Calendar.MONTH) == current.get(Calendar.MONTH)
                && calendar.get(Calendar.DAY_OF_MONTH) == current.get(Calendar.DAY_OF_MONTH);
    }
}
